#!/usr/bin/env python

"""
--- TDSweeper -----------------------------------------
 A free minesweeper clone for console play written in
 Python by Kalle Svensson a.k.a. krftkndl, a proud
 member of Technological Dimensions.
 Visit TD at http://hem1.passagen.se/tech/index.html
--- Description ---------------------------------------
 TDSweeper is a clone of one of the few good things
 Microsoft have given us, minesweeper. It is intended
 for the Linux platform, but might run on other
 platforms too.
 It uses the pyslang module for screen management.
 This can be found on [pyslang address].
--- Copyright and distribution ------------------------
 This program is Copyright (C) 1999 Kalle Svensson.
 It may be distributed under the terms of the GNU GPL
     (GNU General Public License,
     	http://www.gnu.org/copyleft/gpl.html)
 This basically means you can do whatever you like
 with it as long as you make your changes/contributions
 available under the same license. Also, it means that
 there is no warranty at all. (However, I don't think
 it will cause any serious damage to your system...)
 You are not allowed to remove this copyright notice
 either, but why would you want to do that?
 Oh, yeah, almost forgot: IANAL.
--- Contact -------------------------------------------
 Do you have any questions, comments, suggestions or
 improvements? Have you used the code and made
 something absolutely amazing?
 Tell me.
       	krftkndl@hem2.passagen.se
-------------------------------------------------------
"""

__version__ = '1.0.0'

#######################################################
### Modules ###########################################
#######################################################

import whrandom
import time
import sys
import os
import string
import getpass
from slang import *

#######################################################
### Class definitions #################################
#######################################################

class SweepGame:

    def __init__(self):
	SLtt_get_terminfo()
	SLang_init_tty(-1, 0, 0)
	SLsmg_init_smg()
	SLkp_init()
	SLtt_set_color(NORMAL_VIDEO, '', 'lightgray', 'black')
	SLtt_set_color(INVERSE_VIDEO, '', 'black', 'lightgray')
	SLtt_set_color(HIGHLIGHT, '', 'red', 'black')
	SLtt_set_color(FOCUSED_BOX, '', 'black', 'red')
	SLtt_set_color(FLAGGED_BOX, '', 'red', 'black')
	SLtt_set_color(OPENED_BOX, '', 'green', 'black')
	SLtt_set_cursor_visibility(0)
	self.getHiscore()
	self.getDiff()

    def run(self):
	while 1:
	    self.printMenu()
	    input = chr(SLkp_getkey())
	    if input == 'n' or input == '1':
		self.grid = SweepGrid(self,self.diff_x,self.diff_y,self.diff_mines)
		self.grid.draw()
		self.grid.startTimer()
		while 1:
		    inputx = chr(SLkp_getkey())
		    if inputx == 'w':
			self.grid.moveFocus(0, -1)
		    elif inputx == 'a':
			self.grid.moveFocus(-1, 0)
		    elif inputx == 's':
			self.grid.moveFocus(0, 1)
		    elif inputx == 'd':
			self.grid.moveFocus(1, 0)
		    elif inputx == ',':
			self.grid.focusOpen()
		    elif inputx == '.':
			self.grid.focusFlag()
		    elif inputx == 'h':
			self.printHelp()
		    elif inputx == chr(27):    # chr(27) == [ESC]
			self.userQuit()
		    self.grid.draw()
		    if self.grid.ready:
			new_hiscore = 0
			if self.grid.result == 'win':
			    if self.checkHiscore():
				new_hiscore = 1
			self.grid.printEnd(new_hiscore)
			break
		SLkp_getkey()
	    elif input == 'h' or input == '2':
		self.printHelp()
	    elif input == 's' or input == '3':
		self.setDiff()
		self.getDiff()
	    elif input == 'd' or input == '4':
		self.printDiff()
		SLsmg_gotorc(10,3)
		SLsmg_set_color(HIGHLIGHT)
		SLsmg_write_string('Press any key...')
		SLsmg_normal_video()
		SLsmg_refresh()
		SLkp_getkey()
	    elif input == 'v' or input == '5':
		self.printHiscore()
	    elif input == 'q' or input == '6':
		break
	    elif input == chr(27):    # chr(27) == [ESC]
		self.userQuit()

    def getHiscore(self):

	self.hiscores = []

	try:
	    hiscore_file = open('.sweephs')
	    hiscore_line_list = hiscore_file.readlines()
	    hiscore_file.close()
	except IOError:
	    hiscore_file = open('.sweephs','w')
	    hiscore_file.write('')
	    hiscore_file.close()
#	    os.chmod('.sweephs',0666)
	    hiscore_line_list = []

	for hiscore_line in hiscore_line_list:
	    hiscore_list = string.split(hiscore_line,'::')
	    hiscore_list = map(string.strip, hiscore_list)
	    for index in range(0, 4):
		hiscore_list[index] = int(hiscore_list[index])
	    hiscore_tuple = tuple(hiscore_list)
	    self.hiscores.append(apply(SweepScore, hiscore_tuple))

    def printHiscore(self):
	"""Prints the hiscore screen."""

	SLsmg_cls()
	title = 'linesweeper, the minesweeper clone for linux.'

	SLsmg_gotorc(1,1)
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_draw_box(1,1,3,len(title)+4)
	SLsmg_normal_video()

	SLsmg_gotorc(2,3)
	SLsmg_write_string(title)

	SLsmg_gotorc(4,3)
	SLsmg_write_string('Hiscores:')

	curr_row = 6
	for hiscore in self.hiscores:
	    SLsmg_gotorc(curr_row,3)
	    SLsmg_write_string('Rows: '+`hiscore.diff_y`+' Cols: '+`hiscore.diff_x`+' Mines: '+`hiscore.diff_mines`+' Time: '+`hiscore.time`+' (by '+hiscore.name+')')
	    curr_row = curr_row + 1

	SLsmg_gotorc(curr_row+1,3)
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('Press any key...')
	SLsmg_normal_video()

	SLsmg_refresh()
	SLkp_getkey()

    def checkHiscore(self):
	"""Checks if a result is a highscore and saves highscores to file."""

	game_score = SweepScore(self.diff_x, self.diff_y, self.diff_mines, self.grid.gametime)
	diff_in_hiscores = 0
	hiscores_changed = 0

	for hiscore in self.hiscores:
	    if (game_score.diff_x == hiscore.diff_x) and (game_score.diff_y == hiscore.diff_y) and (game_score.diff_mines == hiscore.diff_mines):
		diff_in_hiscores = 1
		old_hiscore_index = self.hiscores.index(hiscore)
	if diff_in_hiscores:
	    if game_score.time < self.hiscores[old_hiscore_index].time:
		self.hiscores[old_hiscore_index] = game_score
		hiscores_changed = 1
	else:
	    self.hiscores.append(game_score)
	    hiscores_changed = 1

	if hiscores_changed:
	    hiscore_file = open('.sweephs','w')
	    for hiscore in self.hiscores:
		hiscore_file.write(`hiscore.diff_x`+'::'+`hiscore.diff_y`+'::'+`hiscore.diff_mines`+'::'+`hiscore.time`+'::'+hiscore.name+'\n')
	    hiscore_file.close()
	    self.getHiscore()

	return hiscores_changed

    def printMenu(self):
	"""Prints the menu screen."""

	SLsmg_cls()
	title = 'linesweeper, the minesweeper clone for linux.'

	SLsmg_gotorc(1,1)
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_draw_box(1,1,3,len(title)+4)
	SLsmg_normal_video()

	SLsmg_gotorc(2,3)
	SLsmg_write_string(title)

	SLsmg_gotorc(4,3)
	SLsmg_write_string('1) ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('N')
	SLsmg_normal_video()
	SLsmg_write_string('ew game.')

	SLsmg_gotorc(5,3)
	SLsmg_write_string('2) ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('H')
	SLsmg_normal_video()
	SLsmg_write_string('elp.')

	SLsmg_gotorc(6,3)
	SLsmg_write_string('3) ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('S')
	SLsmg_normal_video()
	SLsmg_write_string('et difficulty.')

	SLsmg_gotorc(7,3)
	SLsmg_write_string('4) View current ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('d')
	SLsmg_normal_video()
	SLsmg_write_string('ifficulty.')

	SLsmg_gotorc(8,3)
	SLsmg_write_string('5) ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('V')
	SLsmg_normal_video()
	SLsmg_write_string('iew highscores.')

	SLsmg_gotorc(9,3)
	SLsmg_write_string('6) ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('Q')
	SLsmg_normal_video()
	SLsmg_write_string('uit.')

	SLsmg_refresh()


    def printHelp(self):
	"""Prints the help screen."""

	SLsmg_cls()
	title = 'linesweeper, the minesweeper clone for linux.'

	SLsmg_gotorc(1,1)
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_draw_box(1,1,3,len(title)+4)
	SLsmg_normal_video()

	SLsmg_gotorc(2,3)
	SLsmg_write_string(title)

	SLsmg_gotorc(4,3)
	SLsmg_write_string('Keystrokes:')

	SLsmg_gotorc(6,3)
	SLsmg_write_string('w = up')

	SLsmg_gotorc(7,3)
	SLsmg_write_string('a = left')

	SLsmg_gotorc(6,14)
	SLsmg_write_string('s = down')

	SLsmg_gotorc(7,14)
	SLsmg_write_string('d = right')

	SLsmg_gotorc(8,3)
	SLsmg_write_string(', = open')

	SLsmg_gotorc(8,14)
	SLsmg_write_string('. = flag')

	SLsmg_gotorc(10,3)
	SLsmg_write_string('[ESC] = Quit to system')

	SLsmg_gotorc(12,3)
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('Press any key...')
	SLsmg_normal_video()

	SLsmg_refresh()
	SLkp_getkey()


    def getDiff(self):
	"""Gets difficulty settings from a file."""

	try:
	    diff_file = open('.sweepdiff')
	    diff_line = diff_file.readline()
	    diff_file.close()
	except IOError:
	    diff_file = open('.sweepdiff','w')
	    diff_line = '8::8::10'
	    diff_file.write(diff_line)
	    diff_file.close()
#	    os.chmod('.sweepdiff',0666)

	diff_list = string.split(diff_line,'::')
	diff_list = map(int, diff_list)
	self.diff_x, self.diff_y, self.diff_mines = tuple(diff_list)

    def printSetDiffMenu(self):
	"""Prints the menu for setting difficulty."""

	SLsmg_gotorc(10,3)
	SLsmg_write_string('1) ')
	SLsmg_write_string('Set difficulty to ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('b')
	SLsmg_normal_video()
	SLsmg_write_string('eginner. (8x8, 10 mines)')

	SLsmg_gotorc(11,3)
	SLsmg_write_string('2) ')
	SLsmg_write_string('Set difficulty to ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('i')
	SLsmg_normal_video()
	SLsmg_write_string('ntermediate. (16x16, 40 mines)')

	SLsmg_gotorc(12,3)
	SLsmg_write_string('3) ')
	SLsmg_write_string('Set difficulty to ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('a')
	SLsmg_normal_video()
	SLsmg_write_string('dvanced. (16x30, 99 mines)')

	SLsmg_gotorc(13,3)
	SLsmg_write_string('4) ')
	SLsmg_write_string('Set difficulty to ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('c')
	SLsmg_normal_video()
	SLsmg_write_string('ustom...')

	SLsmg_gotorc(14,3)
	SLsmg_write_string('5) ')
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string('S')
	SLsmg_normal_video()
	SLsmg_write_string('ave settings.')

	SLsmg_refresh()


    def printDiff(self):

	SLsmg_cls()
	title = 'linesweeper, the minesweeper clone for linux.'

	SLsmg_gotorc(1,1)
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_draw_box(1,1,3,len(title)+4)
	SLsmg_normal_video()

	SLsmg_gotorc(2,3)
	SLsmg_write_string(title)

	SLsmg_gotorc(4,3)
	SLsmg_write_string('Current settings:')

	SLsmg_gotorc(6,3)
	SLsmg_write_string('Rows: '+`self.diff_y`)

	SLsmg_gotorc(7,3)
	SLsmg_write_string('Cols: '+`self.diff_x`)

	SLsmg_gotorc(8,3)
	SLsmg_write_string('Mines: '+`self.diff_mines`)


    def setDiff(self):
	"""Sets difficulty and saves the settings to a file."""

	diff_changed = 0

	while 1:
	    self.printDiff()
	    self.printSetDiffMenu()
	    input = chr(SLkp_getkey())
	    if input == 'b' or input == '1':
		self.diff_x, self.diff_y, self.diff_mines = 8, 8, 10
		diff_changed = 1
	    elif input == 'i' or input == '2':
		self.diff_x, self.diff_y, self.diff_mines = 16, 16, 40
		diff_changed = 1
	    elif input == 'a' or input == '3':
		self.diff_x, self.diff_y, self.diff_mines = 30, 16, 99
		diff_changed = 1
	    elif input == 'c' or input == '4':
		self.setCustomDiff()
		diff_changed = 1
	    elif input == 's' or input == '5':
		break
	    elif input == chr(27):    # chr(27) == [ESC]
		self.userQuit()

	if diff_changed:
	    diff_file = open('.sweepdiff','w')
	    diff_file.write(`self.diff_x`+'::'+`self.diff_y`+'::'+`self.diff_mines`)
	    diff_file.close()


    def setCustomDiff(self):

	rowstring = ''
	colstring = ''
	minestring = ''

	SLsmg_cls()
	self.printDiff()
	SLsmg_gotorc(10,3)
	SLsmg_write_string('Rows? ')
	SLsmg_refresh()
	while 1:
	    input = chr(SLkp_getkey())
	    if input in string.digits:
		rowstring = rowstring + input
		SLsmg_write_string(input)
		SLsmg_refresh()
	    elif input == chr(13): # Return
		break
	    elif input == chr(27):    # chr(27) == [ESC]
		userQuit()

	SLsmg_gotorc(11,3)
	SLsmg_write_string('Cols? ')
	SLsmg_refresh()
	while 1:
	    input = chr(SLkp_getkey())
	    if input in string.digits:
		colstring = colstring + input
		SLsmg_write_string(input)
		SLsmg_refresh()
	    elif input == chr(13): # Return
		break
	    elif input == chr(27):    # chr(27) == [ESC]
		userQuit()

	SLsmg_gotorc(12,3)
	SLsmg_write_string('Mines? ')
	SLsmg_refresh()
	while 1:
	    input = chr(SLkp_getkey())
	    if input in string.digits:
		minestring = minestring + input
		SLsmg_write_string(input)
		SLsmg_refresh()
	    elif input == chr(13): # Return
		break
	    elif input == chr(27):    # chr(27) == [ESC]
		userQuit()

	self.diff_x = int(colstring)
	self.diff_y = int(rowstring)
	self.diff_mines = int(minestring)


    def finalize(self):
	"""Resets the screen and tty."""

	SLtt_set_cursor_visibility(1) # doesn't work on win32
	SLsmg_reset_smg()
	SLang_reset_tty()
	print
	print "Bye!"
	print


    def userQuit(self):
	"""A function called by pressing [ESC]. Calls finalize(), then sys.exit()"""

	self.finalize()
	sys.exit(0)


class SweepGrid:

    def __init__(self, game, arg_x, arg_y, arg_mines):
	self.game = game
	self.mines = arg_mines
	self.x = arg_x
	self.y = arg_y
	self.xfocus = 0
	self.yfocus = 0
	self.boxes_flagged = 0
	self.mines_flagged = 0
	self.ready = 0
	self.rows = []
	for rowloop in range(0,self.y):
	    self.rows.append(SweepRow(self.game, rowloop, self.x))
	self.initializeMines()

    def boxFlagged(self, is_mine):
	self.boxes_flagged = self.boxes_flagged + 1
	if is_mine:
	    self.mines_flagged = self.mines_flagged + 1
	if (self.mines_flagged == self.mines) and (self.boxes_flagged == self.mines):
	    self.end('win')

    def boxUnflagged(self, is_mine):
	self.boxes_flagged = self.boxes_flagged - 1
	if is_mine:
	    self.mines_flagged = self.mines_flagged - 1
	if (self.mines_flagged == self.mines) and (self.boxes_flagged == self.mines):
	    self.end('win')

    def initializeMines(self):
	mines = 0
	while mines < self.mines:
	    x = whrandom.randint(0,self.x-1)
	    y = whrandom.randint(0,self.y-1)
	    if not self.rows[y].cols[x].is_mine:
		self.rows[y].cols[x].isMine()
		mines = mines + 1

    def moveFocus(self, xmove, ymove):
	new_xfocus = self.xfocus + xmove
	new_yfocus = self.yfocus + ymove
	if new_xfocus >= 0 and new_xfocus < self.x:
	    self.xfocus = new_xfocus
	if new_yfocus >= 0 and new_yfocus < self.y:
	    self.yfocus = new_yfocus

    def focusOpen(self):
	self.rows[self.yfocus].cols[self.xfocus].open()

    def focusFlag(self):
	self.rows[self.yfocus].cols[self.xfocus].toggleFlag()

    def draw(self):
	SLsmg_cls()
	SLsmg_draw_box(0,0,self.y+1,self.x+1)
	SLsmg_gotorc(self.y,0)
	SLsmg_draw_hline(self.x)
	SLsmg_gotorc(0,self.x)
	SLsmg_draw_vline(self.y)
	for row in self.rows:
	    for box in row.cols:
		if (box.x == self.xfocus) and (box.y == self.yfocus):
		    SLsmg_gotorc(box.y, box.x)
		    box.draw(1)
		else:
		    SLsmg_gotorc(box.y, box.x)
		    box.draw()
	SLsmg_gotorc(self.yfocus, self.xfocus)
	SLsmg_refresh()

    def startTimer(self):
	self.starttime = time.time()

    def end(self, result):
	SLsmg_write_string('press h for help.')
	SLsmg_refresh()
	endtime = time.time()
	self.gametime = endtime - self.starttime
	self.gametime = int(self.gametime)
	self.result = result
	self.ready = 1

    def printEnd(self, new_hiscore):
	SLsmg_normal_video()
	SLsmg_gotorc(self.y + 1, 2)
	SLsmg_write_string("You " + self.result + "!")
	SLsmg_gotorc(self.y + 2, 2)
	SLsmg_write_string("It took " + `self.gametime` + " seconds!")
	curr_row = 3
	if new_hiscore:
	    SLsmg_gotorc(self.y + curr_row, 2)
	    SLsmg_write_string("That's a new highscore!")
	    curr_row = curr_row + 1
	SLsmg_gotorc(self.y+curr_row+1, 2)
	SLsmg_set_color(HIGHLIGHT)
	SLsmg_write_string("Press a key...")
	SLsmg_normal_video()
	SLsmg_refresh()


class SweepRow:

    def __init__(self, game, arg_y, xrange):
	self.game = game
	self.cols = []
	self.y = arg_y
	for colloop in range(0,xrange):
	    self.cols.append(SweepBox(self.game, colloop, self.y))


class SweepBox:

    def __init__(self, game, arg_x, arg_y):
	self.game = game
	self.x = arg_x
	self.y = arg_y
	self.opened = 0
	self.flagged = 0
	self.is_mine = 0
	self.neighboring_mines = 0
	self.draw_as = "#"

    def isMine(self):
	self.is_mine = 1

    def open(self):
	if not self.flagged and not self.opened:
	    if self.is_mine:
		self.draw_as = "*"
		self.game.grid.end('lose')
	    else:
		self.checkNeighbors()
		self.opened = 1
		if self.neighboring_mines == 0:
		    self.openNeighbors()

    def toggleFlag(self):
	if not self.opened:
	    if not self.flagged:
		self.game.grid.boxFlagged(self.is_mine)
		self.flagged = 1
		self.draw_as = '!'
	    else:
		self.game.grid.boxUnflagged(self.is_mine)
		self.flagged = 0
		self.draw_as = '#'

    def checkNeighbors(self):
	if self.x-1 >= 0:
	    if self.y-1 >= 0:
		if self.game.grid.rows[self.y-1].cols[self.x-1].is_mine:
		    self.neighboring_mines = self.neighboring_mines + 1
	    if self.game.grid.rows[self.y].cols[self.x-1].is_mine:
		self.neighboring_mines = self.neighboring_mines + 1
	    if self.y+1 < self.game.grid.y:
		if self.game.grid.rows[self.y+1].cols[self.x-1].is_mine:
		    self.neighboring_mines = self.neighboring_mines + 1
	if self.y-1 >= 0:
	    if self.game.grid.rows[self.y-1].cols[self.x].is_mine:
		self.neighboring_mines = self.neighboring_mines + 1
	if self.y+1 < self.game.grid.y:
	    if self.game.grid.rows[self.y+1].cols[self.x].is_mine:
		self.neighboring_mines = self.neighboring_mines + 1
	if self.x+1 < self.game.grid.x:
	    if self.y-1 >= 0:
		if self.game.grid.rows[self.y-1].cols[self.x+1].is_mine:
		    self.neighboring_mines = self.neighboring_mines + 1
	    if self.game.grid.rows[self.y].cols[self.x+1].is_mine:
		self.neighboring_mines = self.neighboring_mines + 1
	    if self.y+1 < self.game.grid.y:
		if self.game.grid.rows[self.y+1].cols[self.x+1].is_mine:
		    self.neighboring_mines = self.neighboring_mines + 1
	if self.neighboring_mines:
	    self.draw_as = `self.neighboring_mines`
	else:
	    self.draw_as = ' '

    def openNeighbors(self):
	if self.x-1 >= 0:
	    if self.y-1 >= 0:
		self.game.grid.rows[self.y-1].cols[self.x-1].open()
	    self.game.grid.rows[self.y].cols[self.x-1].open()
	    if self.y+1 < self.game.grid.y:
		self.game.grid.rows[self.y+1].cols[self.x-1].open()
	if self.y-1 >= 0:
	    self.game.grid.rows[self.y-1].cols[self.x].open()
	if self.y+1 < self.game.grid.y:
	    self.game.grid.rows[self.y+1].cols[self.x].open()
	if self.x+1 < self.game.grid.x:
	    if self.y-1 >= 0:
		self.game.grid.rows[self.y-1].cols[self.x+1].open()
	    self.game.grid.rows[self.y].cols[self.x+1].open()
	    if self.y+1 < self.game.grid.y:
		self.game.grid.rows[self.y+1].cols[self.x+1].open()

    def draw(self, focus = 0):
	if focus:
	    SLsmg_set_color(FOCUSED_BOX)
	elif self.flagged:
	    SLsmg_set_color(FLAGGED_BOX)
	elif self.opened:
	    SLsmg_set_color(OPENED_BOX)
	SLsmg_write_string(self.draw_as)
	SLsmg_normal_video()

class SweepScore:
    def __init__(self, diff_x, diff_y, diff_mines, time, name = ''):
	self.diff_x, self.diff_y, self.diff_mines, self.time = diff_x, diff_y, diff_mines, time
	if name:
	    self.name = name
	else:
	    self.name = getpass.getuser()

#######################################################
### Macros ############################################
#######################################################

NORMAL_VIDEO = 0
INVERSE_VIDEO = 1
HIGHLIGHT = 2
FOCUSED_BOX = 3
FLAGGED_BOX = 4
OPENED_BOX = 5

#######################################################
### Main ##############################################
#######################################################

os.chdir(os.environ['HOME'])
fisk = SweepGame()
fisk.run()
fisk.finalize()
